﻿#ifndef C2DPUNKT_H
#define C2DPUNKT_H

//------ ###Aufgabe 9.2.2, 2. -----------------------------------------------------


// Diese Aufgabenstellung ist etwas vage.
// Vorstellbar sind aber z.B. die folgenden Anforderungen:
// - Einen C2DPunkt mit einzelnen Koordinaten sowie einem anderen
//   C2DPunkt zu initialisieren.
// - Jede Koordinate einzeln zu setzen und zu lesen
// - Beide Koordinaten auf einen anderen C2DPunkt zu setzen
// - Den Punkt als String darzustellen

class C2DPunkt{
  double x,y;
 public:
  // Lösung 9.1.5, 3.
  C2DPunkt(const double& x_=0, const double& y_=0) {x=x_; y=y_;}
  C2DPunkt(const C2DPunkt& p) {x=p.x; y=p.y;}

  // Lösung 9.2.2, 2. mit Elementinitialisierern
  // C2DPunkt(const double& x_=0, const double& y_=0):x(x_), y(y_) {  }
  // C2DPunkt(const C2DPunkt& p):x(p.x), y(p.y) {  }

  // Später wird gezeigt, dass dieser zweite Konstruktor ein
  // Kopierkonstruktor ist, der automatisch erzeugt wird und
  // deswegen nicht definiert werden muß.

  // Viele Konstruktoren haben wie hier einen leeren Anweisungsteil
  void setzeX(const double& x_) { x=x_; }
  void setzeY(const double& y_) { y=y_; }
  double X() const { return x; }
  double Y() const { return y; }
  void setze(const double& x_, const double& y_) { x=x_; y=y_; }
  void setze(const C2DPunkt& p) { x=p.x; y=p.y; }
  // Später werden wir sehen, dass diese Funktion nicht notwendig ist,
  // da der vom Compiler erzeugte Zuweisungsoperator denselben Effekt hat.

  std::string toStr() const
  {
    return "("+std::to_string(x) + "|" + std::to_string(y)+")";
  } // { z. B. (2,345|3,45678) }
};

#ifdef SIMPLEUNITTESTS_H__ 
// Dieses Makro wird durch ein #include "SimpleUnitTests.h" definiert

void Unittests_C2DPunkt()
{ // teste alle public Elementfunktionen und Konstruktoren
  using namespace rk1;
  C2DPunkt p1(2, 3);
  Assert::AreEqual(2.0, p1.X(), "p1.x==2");

  Assert::AreEqual(3.0, p1.Y(), "p1.y==3");

  C2DPunkt p2(2);
  Assert::AreEqual(2.0, p2.X(), "p2.x==2");
  Assert::AreEqual(0.0, p2.Y(), "p2.y==0");

  C2DPunkt p3;
  Assert::AreEqual(0.0, p3.X(), "p3.x==0");
  Assert::AreEqual(0.0, p3.Y(), "p3.y==0");

  C2DPunkt p4(p1);
  Assert::AreEqual(p1.X(), p4.X(), "p3.x==p1.x");
  Assert::AreEqual(p1.Y(), p4.Y(), "p3.y==p1.y");

  p3.setze(5, 6);
  Assert::AreEqual(5.0, p3.X(), "p1.x==5");
  Assert::AreEqual(6.0, p3.Y(), "p1.y==6");

  p3.setze(p1);
  Assert::AreEqual(p1.X(), p3.X(), "p3.x==p1.x");
  Assert::AreEqual(p1.Y(), p3.Y(), "p3.y==p1.y");

  p3.setzeX(8);
  Assert::AreEqual(8.0, p3.X(), "p3.x==p1.x");
  p3.setzeY(9);
  Assert::AreEqual(9.0, p3.Y(), "p3.y==p1.y");
}
#endif // #ifdef SIMPLEUNITTESTS_H__ 


#endif // C2DPUNKT_H

