﻿// *********** Kap. 16 Lösungen Kapitel 16. STL-Algorithmen

#include <string>
#include <sstream>   // ??? für Exception Text
#include <fstream>
#include <stdexcept> // für range_error
#include <vector>
#include <algorithm>
#include <set>
#include <map>
#include <iterator>
#include <utility>
#include <tuple>
#include <functional>
#include <valarray>
#include <complex>
#include <cmath>


// *********** Kap. 16 Aufgabe 16.1 Iteratoren

namespace N_Iteratoren 
{


  using namespace System::Windows::Forms;

  // --------------- Aufgabe 1 -------------------------------

  // a)
  template<typename T, typename Iterator>
  void writeToFile(Iterator first, Iterator last, std::string fn)
  {
    std::ofstream f(fn);
    std::copy(first, last, std::ostream_iterator<T>(f, "\n"));
  }

  // b)
  template<typename T>
  void copyFile(std::string fn1, std::string fn2)
  {
    std::ifstream f1(fn1);
    std::ofstream f2(fn2);
    std::copy(std::istream_iterator<T>(f1),
      std::istream_iterator<T>(), std::ostream_iterator<T>(f2, "\n"));
  }

  // c)
  template<typename T>
  void sortFile(std::string fn1, std::string fn2)
  {
    std::ifstream f1(fn1);
    std::vector<T> v;
    std::copy(std::istream_iterator<T>(f1),
      std::istream_iterator<T>(), std::back_inserter(v));
    std::sort(v.begin(), v.end());
    std::ofstream f2(fn2);
    std::copy(v.begin(), v.end(), std::ostream_iterator<T>(f2, "\n"));
  }


  // d)
  template<typename T>
  void showFile(TextBox^ tb, std::string fn)
  {
    tb->AppendText("Dateiname:" + rk1::to_String(fn) + "\r\n");
    std::ifstream f(fn);
    T x;
    while (f >> x)
    {
      tb->AppendText(rk1::to_String(std::to_string(x) + "\r\n"));
    }
  }

  // e)
  struct Bruch {
    Bruch() :z(0), n(1) {}
    Bruch(int z_, int n_) :z(z_), n(n_) {}
    int z, n; // z: Zähler, n: Nenner
  };

  std::ostream& operator<<(std::ostream& f, const Bruch& b)
  {
    return f << b.z << "|" << b.n;
  }

  std::istream& operator >> (std::istream& f, Bruch& b)
  {
    char Bruchstrich;
    f >> b.z >> Bruchstrich >> b.n;
    return f;
  }

  void Test_a_bis_d(TextBox^ tb)
  {
    const std::string TestDir = "c:\\Test\\";
    // a), e)
    int as[3] = { 1,2,3 };                         // sortiert
    writeToFile<int>(as, as + 3, TestDir + "s.txt");
    int au[3] = { 1,3,2 };                         // unsortiert
    writeToFile<int>(au, au + 3, TestDir + "u.txt");
    int a0[1];
    writeToFile<int>(a0, a0, TestDir + "0.txt");   // 0 Elemente
    int a1[1] = { 1 };
    writeToFile<int>(a1, a1 + 1, TestDir + "1.txt"); // 1 Element


    // b), e)
    copyFile<int>(TestDir + "s.txt", TestDir + "sc.txt");
    copyFile<int>(TestDir + "u.txt", TestDir + "uc.txt");
    copyFile<int>(TestDir + "0.txt", TestDir + "0c.txt");
    copyFile<int>(TestDir + "1.txt", TestDir + "1c.txt");

    // c), e)
    sortFile<int>(TestDir + "s.txt", TestDir + "s_sor.txt");
    sortFile<int>(TestDir + "u.txt", TestDir + "u_sor.txt");
    sortFile<int>(TestDir + "0.txt", TestDir + "0_sor.txt");
    sortFile<int>(TestDir + "1.txt", TestDir + "1_sor.txt");


    // d)
    showFile<int>(tb, TestDir + "u.txt");
    showFile<int>(tb, TestDir + "0.txt");
    showFile<int>(tb, TestDir + "1.txt");

    showFile<int>(tb, TestDir + "s_sor.txt");
    showFile<int>(tb, TestDir + "u_sor.txt");
    showFile<int>(tb, TestDir + "0_sor.txt");
    showFile<int>(tb, TestDir + "1_sor.txt");

    // f)
    std::vector<Bruch> vb = { Bruch(1, 2),Bruch(1, 3), Bruch(1, 4) };
    writeToFile<N_Iteratoren::Bruch>(vb.begin(), vb.end(), TestDir + "b.txt");
  }


  // --------------- Aufgabe 2 -------------------------------

  /* Zur Lösung dieser Aufgabe müssen der Klasse Array
  lediglich diese Definitonen hinzugefügt werden:

  typedef T*             iterator;
  iterator begin() { return elems; }
  iterator end() { return elems+N; }
  */

  // #include <stdexcept> // für range_error
  // #include <sstream>   // für Exception Text


  template<class T, unsigned int N>
  class Array {
  public:
    T elems[N];    // fixed-size array of elements of type T

                   // kein Konstruktor, damit ein Array wie ein C-Array
                   // initialisiert werden kann
  public:
    // type definitions
    typedef T              value_type;
    typedef T&             reference;
    typedef const T&       const_reference;
    typedef std::size_t    size_type;

    // operator[]
    reference operator[](unsigned int i)
    {
      if (i >= N) // bzw. (i<0 || i>=N) bei int i;
      {
        std::ostringstream msg;
        msg << "Array index out of range: " << i << " > " << (N - 1);
        throw std::out_of_range(msg.str());
      }
      return elems[i];
    }

    // ohne diese Variante ist der Indexoperator für konstante
    // Arrays wie
    //    const Array<int, n> a={1,2,3};
    // nicht definiert:
    const_reference operator[](unsigned int i) const
    {
      if (i >= N) // bzw. (i<0 || i>=N) bei int i;
      {
        std::ostringstream msg;
        msg << "Array index out of range: " << i << " > " << (N - 1);
        throw std::out_of_range(msg.str());
      }
      return elems[i];
    }

    // size is constant
    static size_type size() { return N; }

    // Ergänzungen:

    typedef T*             iterator;
    iterator begin() { return elems; }
    iterator end() { return elems + N; }

    // Die nächsten drei Definitionen ermöglichen den Aufruf
    // von STL-Algorithmen mit konstanten Array.
    typedef const T*       const_iterator;
    const_iterator begin() const { return elems; }
    const_iterator end() const { return elems + N; }
  };


  void test_Array()
  {
    Array<int, 5> a = { 1,9,-1, 6,2 };
    std::sort(a.begin(), a.end());
    {
      const int n = 100;
      Array<int, n> a;
      for (int i = 0; i < n; i++) a[i] = n - i;
      std::sort(a.begin(), a.end());

    }
  }


  // --------------- Aufgabe 3 -------------------------------

  void Vektor_mit_Konstruktor_füllen()
  {
    std::ifstream f("s.txt");
    std::istream_iterator<int> beg1(f);
    std::istream_iterator<int> end1;
    std::vector<int> v1(beg1, end1);

    std::ifstream fb("b.txt");
    std::istream_iterator<Bruch> beg2(fb);
    std::istream_iterator<Bruch> end2;
    std::vector<Bruch> vb(beg2, end2);
  }
} // end of namespace N_Iteratoren 




// *********** Kap. 16 Aufgabe 16.13 STL-Algorithmen
namespace N_STL_Algorithmen
{
  using namespace System::Windows::Forms;




  const std::string TestDir = "c:\\Test\\";

  //------ ###Aufgabe 1 -----------------------------------------------------
  namespace Meine_STL
  {
    //------ Aufgabe 1 -----------------------------------------------------

    // Die Algorithmen der STL sind mit dem hier auskommentierten 
    // und durch int ersetzten difference_type etwas universeller.

    template <class InputIterator, class T>
    // typename iterator_traits<InputIterator>//::difference_type
    int count(InputIterator first, InputIterator last, const T& value)
    {
      // typename iterator_traits<InputIterator>::difference_type 
      int n = 0;
      for (; first != last; ++first)
        if (*first == value)
          ++n;
      return n;
    }

    template <class InputIterator, class Predicate>
    // typename iterator_traits<InputIterator>::difference_type
    int count_if(InputIterator first, InputIterator last, Predicate pred)
    {
      // typename iterator_traits<InputIterator>::difference_type 
      int n = 0;
      for (; first != last; ++first)
        if (pred(*first))
          ++n;
      return n;
    }

    /* **** alte Versionen (nicht mehr im C++-Standard, aber aus STL.zip
    template <class InputIterator, class T, class Size>
    void count(InputIterator first, InputIterator last, const T& value,
    Size& n) {
    while (first != last)
    if (*first++ == value) ++n;
    }

    template <class InputIterator, class Predicate, class Size>
    void count_if(InputIterator first, InputIterator last, Predicate pred,
    Size& n) {
    while (first != last)
    if (pred(*first++)) ++n;
    }
    */

    // In algo.h aus "ftp://butler.hpl.hp.com/stl/stl.zip" sind
    // die Algorithmen min_element usw. folgendermaßen implementiert:

    template <class ForwardIterator>
    ForwardIterator min_element(ForwardIterator first, ForwardIterator last)
    {
      if (first == last) return first;
      ForwardIterator result = first;
      while (++first != last)
        if (*first < *result) result = first;
      return result;
    }

    template <class ForwardIterator, class Compare>
    ForwardIterator min_element(ForwardIterator first, ForwardIterator last,
      Compare comp)
    {
      if (first == last) return first;
      ForwardIterator result = first;
      while (++first != last)
        if (comp(*first, *result)) result = first;
      return result;
    }

  } // end of namespace Meine_STL

  void Aufgabe_1()
  {
    std::vector<int> v = { 1,5,2,4,4 };
    std::set<double> s = { 1,5,2,4,4 };
    int i = Meine_STL::count(v.begin(), v.end(), 1);
    int j = Meine_STL::count_if(v.begin(), v.end(), [](int i) {return i % 2 == 1; });
    int k = *Meine_STL::min_element(v.begin(), v.end());
  }

  //------ Aufgabe 2 -----------------------------------------------------

  template <typename Container>
  typename Container::value_type Median(Container c)
  {
    nth_element(c.begin(), c.begin() + c.size() / 2, c.end());
    return *(c.begin() + c.size() / 2);
  }

  void test_Median()
  {
    // ???
    std::vector<double> v1 = { 1, 2, 4, 4, 4, 5, 15 };
    double md1 = Median(v1); // 4

    std::vector<int> v2{ 5, 6, 4, 3, 2, 6, 7, 9, 3 };
    double md2 = Median(v2); // 5
  }


// *********** Kap. 16 Aufgabe 16.14.2 Valarrays

  void erzeugeTestdaten(std::valarray<double>& x, std::valarray<double>& y, 
    int n, double a_, double b_)
  {
    // Aufgabe 4.5.2, 1
    // Aufgabe 4.5.2, 2 a): rnd=0;
    // Aufgabe 4.5.2, 2 b): rnd>0;
    double rnd = 5; // mit positivem Wert ergibt sich eine Streuung
    for (int i = 0; i < n; i++)
    { // Testwerte erzeugen
      x[i] = i;
      y[i] = a_*x[i] + b_ + rnd*((10 - 1.0) / 2 - rand() % 10);
    }
  }


  // Die folgenden Variablen nur deshalb global, da
  // die Teilaufgaben in eigenen Funktionen behandelt
  // werden:
  double sxy, sx, sy, sxx, ym, xm;
  const int n = 100; // n=2; n=100; n=10000;
  std::valarray<double> x(n), y(n);

  void kleinste_Quadrate_mit_valarrays()
  { // Aufgabe a)
    sxy = (x*y).sum();
    sx = x.sum();
    sy = y.sum();
    sxx = (x*x).sum();
    xm = sx / n;
    ym = sy / n;
  }

  void kleinste_Quadrate_mit_Arrays()
  { // Aufgabe b)
    sxy = 0;
    for (int i = 0; i < n; i++) sxy = sxy + x[i] * y[i];
    sx = 0;
    for (int i = 0; i < n; i++) sx = sx + x[i];
    sy = 0;
    for (int i = 0; i < n; i++) sy = sy + y[i];
    sxx = 0;
    for (int i = 0; i < n; i++) sxx = sxx + x[i] * x[i];
    xm = sx / n;
    ym = sy / n;
  }


// *********** Kap. 16 Aufgabe 16.14.4 Komplexe Zahlen

  String^ ComplexToStr(std::complex<double> c)
  {
    return real(c).ToString() + "+" + imag(c).ToString() + "i";
  }

  void QuadGlClick(TextBox^ tb)
  {
    std::complex<double> a(2, 3), b(4, -5), c(-7, 8), x1, x2;
    //  complex<double> a(2,0), b(4,0), c(-7,0), x1, x2;
    x1 = (-b + sqrt(b*b - 4.0*a*c)) / (2.0*a);
    x2 = (-b - sqrt(b*b - 4.0*a*c)) / (2.0*a);
    tb->AppendText(ComplexToStr(x1) + "\r\n");
    tb->AppendText(ComplexToStr(x2) + "\r\n");
    // Probe:
    tb->AppendText(ComplexToStr(a*x1*x1 + b*x1 + c) + "\r\n");
    tb->AppendText(ComplexToStr(a*x2*x2 + b*x2 + c) + "\r\n");
  }

  // Mit typedef lassen sich die Datentypen etwas einfacher schreiben:

  // --------------- Aufgabe 2 -------------------------------

  void EinheitswurzelnClick(TextBox^ tb)
  {
    const int n = 100;

    const double pi = 3.14159265358979323846;

    std::complex<double> w(std::cos(2 * pi / n), std::sin(2 * pi / n));
    for (int i = 0; i < n; i++)
    {
      std::complex<double> z = pow(w, i); // Einheitswurzel
                                     // z^n muss etwa 1 sein:
      tb->AppendText(ComplexToStr(pow(z, n)) + "\r\n");
    }
  }



} // end of namespace N_STL_Algorithmen

