﻿// *********** Kap. 18 Lösungen Kapitel 18. Smart Pointer


#include <memory>
// *********** Kap. 18 Aufgabe 18.3

  // Noch als Aufgabe bei unique_ptr:
  // In der Klasse C_Lsg soll beim Aufruf von init eine Exception auftreten 
  // können. Damit in diesem Fall kein Speicherleck auftritt, wurde für einen
  // älteren Compiler, der noch keine unique_ptr kennt, diese Variante mit try-catch 
  // gewählt.
  // Überarbeiten Sie diese Variante mit unique_ptr so, dass kein try-catch mehr notwendig 
  // ist und bei einer Exception in init trotzdem kein Speicherleck mehr auftritt:
namespace N_Loesungen_Smart_Pointer
{

  using namespace System::Windows::Forms;


  using std::unique_ptr;
  using std::make_unique;
  using std::shared_ptr;
  using std::make_shared;
  // --------------- Aufgabe 1. ------------------------------


  void Aufgabe_1()
  {
    // a)
    { // Ein vector<unique_ptr> kann nur mit Move-Semantik gefüllt werden,
      // nicht mit Kopier-Semantik
      std::vector<unique_ptr<int>> vu;
      unique_ptr<int> up = unique_ptr<int>(new int(2));

      // Kopiersemantik geht mit unique_ptr nicht:
      // vu.push_back(up);    
      // vu.emplace_back(up); 

      // Move-Semantik geht mit unique_ptr:
      vu.push_back(unique_ptr<int>(new int(2)));
      vu.emplace_back(unique_ptr<int>(new int(2)));

      // Move-Semantik geht mit make_unique:
      vu.push_back(make_unique<int>(2));
      vu.emplace_back(make_unique<int>(2));

      // b)
      // Ein vector<shared_ptr> kann mit Move- und KopierSemantik 
      // gefüllt werden:
      std::vector<shared_ptr<int>> vs;
      shared_ptr<int> sp = shared_ptr<int>(new int(2));

      // Kopiersemantik geht mit shared_ptr:
      vs.push_back(sp);
      vs.emplace_back(sp);

      // Move-Semantik geht auch:
      vs.push_back(make_shared<int>(2)); // Das geht 
      vs.emplace_back(make_shared<int>(3)); // Das geht 
    }
    // c)
    {
      std::vector<shared_ptr<int>> v;
      for (int i = 0; i < 10; i++)
        v.push_back(make_shared<int>(10 - i));
      std::sort(v.begin(), v.end(), [](shared_ptr<int> x, shared_ptr<int> y)
      {return std::greater<int>()(*x, *y); });
    }
    {
      std::vector<unique_ptr<int>> v;
      for (int i = 0; i < 10; i++)
        v.push_back(make_unique<int>(10 - i));
      // std::sort(v.begin(), v.end(), [](unique_ptr<int> x, 
      //             unique_ptr<int> y) {return greater<int>()(*x, *y); });
    }
  }

  // --------------- Aufgabe 2. ------------------------------

  class C_alt {
    int* p;
  public:
    C_alt(int n) :p(new int(n)) {}
    C_alt(const C_alt& x)
    {
      // ...
    }
    C_alt& operator=(const C_alt& x)
    {
      // ...
    }
    ~C_alt() { delete p; }
  };

  class C_neu_SP {
    std::shared_ptr<int> p;
  public:
    C_neu_SP(int n) :p(make_shared<int>(n)) {}
    // Alle diese folgenden Funktionen sind überflüssig, 
    // da der Datentyp von p eine Klasse ist:
    //   C_neu(const C_neu& x) // der Kopierkonstruktor ist überflüssig
    //   C_neu& operator=(const C_neu& x) // ebenfalls ist überflüssig
    //   ~C_alt() { delete p; } // der Destruktor ist überflüssig
  };

  class C_neu_UP {
    std::unique_ptr<int> p;
  public:
    C_neu_UP(int n) :p(make_unique<int>(n)) {}
    // Alle diese folgenden Funktionen sind überflüssig, 
    // da der Datentyp von p eine Klasse ist:
    //   C_neu(const C_neu& x) // der Kopierkonstruktor ist überflüssig
    //   C_neu& operator=(const C_neu& x) // ebenfalls ist überflüssig
    //   ~C_alt() { delete p; } // der Destruktor ist überflüssig
  };

  void Aufgabe_2()
  {
    C_neu_SP s1(17);
    C_neu_SP s2 = s1;
    s2 = s1;

    // Kopieren geht mit C_neu_UP nicht
    C_neu_UP c1(17);
    // C_neu_UP c2 = c1;
    // c2 = c1;
  }


  // --------------- Aufgabe 3. ------------------------------

  // a)
  class C_neu {
    // std::unique_ptr<int> a;
    // std::unique_ptr<int> b;
    std::shared_ptr<int> a;
    std::shared_ptr<int> b;
  // b) Mit einem unique_ptr können Variablen des Typs C_neu
  //    nicht kopiert werden.      
#if VS_VERSION >= 2015   
    void init() noexcept(false) {};
#else
    void init() {};
#endif
  public:
    // C_neu() : a(std::make_unique<int>(int())), b(new int())
    C_neu() : a(std::make_shared<int>(int())), b(new int())
    {
      // Da a und b jetzt Klassen sind, wird für sie bei
      // einer Exception der Destruktor aufgerufen.
      init();
    }
  };

  void Aufgabe_3()
  {
    C_neu c;
  }

  // c) Da C Zeiger enthält, fehlen der Destruktor, 
  //    Kopierkonstruktor und der Zuweisungsoperator. 

  // d) C_neu enthält keine Zeiger. Der Destruktor/Kopier-
  //    konstruktor/Zuweisungsoperator von C_neu ruft die 
  //    Destruktoren/Kopierkonstruktoren/Zuweisungsoperatoren 
  //    aller Datenelemente auf. Deswegen sind diese 
  //    Funktionen in C_neu nicht notwendig.

  // --------------- Aufgabe 4. ------------------------------
  class C1 {
  public:
    C1() = default;
    C1(const C1&) {};
  };
  void Aufgabe_4a()
  { // http://stackoverflow.com/questions/701456/what-are-potential-dangers-when-using-boostshared-ptr
    C1* p1 = new C1();
    shared_ptr<C1> s1(p1);
    shared_ptr<C1> s2(p1);
    /* Lösung:
    Da zwei shared_ptr auf dieselbe dynamische Variable zeigen
    wird für diese zwei mal delete aufgerufen.

    Mit make_shared kann dieser Fehler nicht auftreten.
    */
  }

  void Aufgabe_4b()
  { // http://stackoverflow.com/questions/701456/what-are-potential-dangers-when-using-boostshared-ptr
    shared_ptr<C1> s1(new C1());
    shared_ptr<C1> s2(s1.get());
    /*
    Da zwei shared_ptr auf dieselbe dynamische Variable zeigen
    wird für diese zwei mal delete aufgerufen.
    */
  }

  void Aufgabe_4c(TextBox^ tb)
  { // http://stackoverflow.com/questions/701456/what-are-potential-dangers-when-using-boostshared-ptr
    int* p = new int(17);
    {
      shared_ptr<int> s1(p);
    }
    tb->AppendText((*p).ToString() + "\r\n");
      /*
      Die dynamische Variable *p wird im Block freigegeben. Der anschließende Zugriff
      ist nicht definiert.
      */
  }

  // --------------- Aufgabe 5. ------------------------------

  struct Node1
  {
    shared_ptr<Node1> next;
  };

  void Z_1(TextBox^ tb)
  {
    Node1* n1 = new Node1;
    Node1* n2 = new Node1;
    shared_ptr<Node1> first(n1);
    first->next = shared_ptr<Node1>(n2);
    tb->AppendText("first.uc=" + first.use_count().ToString() + " first.next.uc=" + first.get()->next.use_count().ToString() +"\r\n");
    n1->next = first; // Zyklus
    tb->AppendText("first.uc=" + first.use_count().ToString() + " first.next.uc=" + first.get()->next.use_count().ToString() + "\r\n");
    // first.uc=2 first.next.uc=2
  } // memory leak


    // --------------- Aufgabe 6. ------------------------------

  void SP_WParameter(std::shared_ptr<int> sp, std::shared_ptr<int>& out)
  {
    sp.get();
    int uc4 = sp.use_count(); // uc4=3
    int uc5 = out.use_count(); // uc5=3
    std::shared_ptr<int> sp1 = sp;
    int uc6 = out.use_count(); // uc6=4
  }

  void test_SP_Parameter()
  {
    std::shared_ptr<int> s1 = std::make_shared<int>(3);
    int uc1 = s1.use_count(); // uc1=1
    std::shared_ptr<int> s2 = s1;
    int uc2 = s1.use_count(); // uc2=2
    SP_WParameter(s1, s1);
    int uc3 = s1.use_count(); // uc3=2
  }


  namespace N_Meinstring_smart_pointer // ???? Aufgabe nur in den Kursunterlagen, nicht im Buch 
  {
    class MeinString {
      char* s; // Zeiger auf nullterminierten String
      int n; // Länge des Strings
    public:
      MeinString(const char* p) // 1
      {
        n = strlen(p);
        s = new char[n + 1];
#pragma warning( push )  
#pragma warning(disable : 4996) //_CRT_SECURE_NO_WARNINGS
        strcpy(s, p);
#pragma warning( pop )   
      };

      MeinString(const MeinString& x) // Kopierkonstruktor
      {
        n = x.n;            // 1
        s = new char[n + 1];  // 2
#pragma warning( push )  
#pragma warning(disable : 4996) //_CRT_SECURE_NO_WARNINGS
        strcpy(s, x.s);    // 3
#pragma warning( pop )   
      };//dieser Konstruktor stellt die Klasseninvariante her

      MeinString& operator=(const MeinString& x)
      {
        if (this != &x)        // a
        {
          delete[] s;          // b
          n = x.n;             // 1
          s = new char[n + 1]; // 2
#pragma warning( push )  
#pragma warning(disable : 4996) //_CRT_SECURE_NO_WARNINGS
          strcpy(s, x.s);      // 3
#pragma warning( pop )   
        }
        return *this;            // c
      };
      ~MeinString() { delete[] s; }
    };

    class MeinString_smart_pointer {
      // std::shared_ptr<char> s = std::make_shared<char>();
      // char* s; // Zeiger auf nullterminierten String
      std::shared_ptr<char[]> s{ new char[1]{'\0'} };
      int n; // Länge des Strings
    public:
      MeinString_smart_pointer(const char* p) // 1
      {
        n = strlen(p);
        s.reset(new char[n + 1]);
#pragma warning( push )  
#pragma warning(disable : 4996) //_CRT_SECURE_NO_WARNINGS
        strcpy(s.get(), p);
#pragma warning( pop )   
      };
      // ~MeinString() { delete[] s; } nicht notwendig
    };
    void testOhneZuweisungsOp()
    {
      {
        using namespace N_Meinstring_smart_pointer;
        MeinString_smart_pointer  s1("abc"), s2(s1), s3("31234");
        s1 = s3;
      }
      MeinString s1("abc"), s2("31234");
      s1 = s2;
      // MeinString s3(s1);
    }
  }



  void test_Aufgaben()
  {
    Aufgabe_1();
    Aufgabe_2();
    Aufgabe_3();
    // Aufgabe 4 führt zu einem Programmabbruch
    // Aufgabe_4a();
    // Aufgabe_4b();
    // Aufgabe_4c();
    // Z_1();
    test_SP_Parameter();
  }

} // end of namespace N_Loesungen_Smart_Pointer



