﻿// *********** Kap. 8 Lösungen Kapitel 8. Überladene Funktionen und Operatoren



#include <cmath> // für pow bei Bruchrechnung
#include <string>
#include <sstream>
#include <iostream>
#include <cstdlib>

namespace N_Loesungen_Ueberladene_Funktionen_und_Operatoren 
{
  using namespace System::Windows::Forms;

// *********** Kap. 8 Aufgabe 8.2

  // --------------- Aufgabe 1 -------------------------------

  // a)
  int fkt(double d) { return 0; }; // ursprüngliche Funktion
  int fkt(int i) { return 1; }; // später hinzugefügt
  int x = fkt(42);

  // Vor der Ergänzung wird die double-Variante aufgerufen, 
  // danach die int-Variante, d.h. x erhält vorher den Wert 0
  // und danach den Wert 1.

  // b) 
  // Für Parameter eines Klassentyps ohne impliziten Konversionsoperator
  // findet keine Konversion statt.    

  // --------------- Aufgabe 2 -------------------------------
  // a)
  void f(std::string s, std::string format) {};
  void f(std::string s) {};
  void f() {};
  // Die Funktionen unter a) können durch eine Funktion 
  // mit Default-Argumenten ersetzt werden:

  void f1(std::string s = "", std::string format = "") {}

  // b)
  void g(int n) {};
  void g(double d) {};
  void g(double d, std::string s) {};

  // Diese Funktionen können nicht durch eine Funktion mit
  // Default-Argumenten ersetzt werden, da die Parameter-Typen
  // verschieden sind.

  // c) 

  namespace N_InitBruch {

    struct Bruch {
      int z, n; // z: Zähler, n: Nenner
    };

    Bruch initBruch(int z, int n)
    {
      return { z,n };
    }

    Bruch initBruch(int z)
    {
      return { z,1 };
    }

    Bruch initBruch()
    {
      return { 1,1 };
    }

    // Variante mit Default-Argumenten:

    Bruch initBruch_(int z = 1, int n = 1) // mit einer Klasse Bruch: Konstruktor
    { // InitBruch(x): x/1
      return { z,n };
    } // 
  } // end of namespace N_InitBruch

  // d) 
  // Default-Argmente sind meist die bessere Alternative, da
  // nur eine einzige Funktion geschrieben und gewartet werden muss.

  // --------------- Aufgabe 3 -------------------------------

  float sqrt(float x) { return 0; };
  double sqrt(double x) { return 0; };
  long double sqrt(long double x) { return 0; };

  void test_sqrt()
  {
    // double x1 = sqrt(1); // mehrdeutig
    double x2 = sqrt(1.0);
    double x3 = sqrt(1.0l); // ein long double Argument
  }

// *********** Kap. 8 Aufgabe 8.3.1

  // --------------- Aufgabe 1 -------------------------------

  // Der Operator ^ kann nur für selbstdefinierte Datentypen überladen 
  // werden. Da '-' eine höhere Priorität als '^' hat, wird x^n-1 als 
  // x^(n-1) ausgewertet. Üblicherweise ist das aber (x^n)-1


  // --------------- Aufgabe 2 -------------------------------

  struct Bruch {
    int z, n; // z: Zähler, n: Nenner
  };


  // a) ==, !=

  inline bool operator==(const Bruch& q1, const Bruch& q2)
  {
    return q1.z*q2.n == q2.z*q1.n;
  }

  // Die folgenden Anweisungen wurden mit leichten Änderungen aus utility.h
  // übernommen. Die Definitionen aus utility.h gehören zum C++-Standard:

  inline bool operator!=(const Bruch& x, const Bruch& y)
  {
    return !(x == y);
  }

  // b) <, <=, >, >=

  inline bool operator<(const Bruch& q1, const Bruch& q2)
  {
    return q1.z*q2.n < q2.z*q1.n;
  }

  inline bool operator>(const Bruch& x, const Bruch& y)
  {
    return y < x;
  }

  inline bool operator<=(const Bruch& x, const Bruch& y)
  {
    return !(y < x);
  }

  inline bool operator>=(const Bruch& x, const Bruch& y)
  {
    return !(x < y);
  }

  // c) +, -, *, /, skalare Multiplikation mit kürzen (-) == *(-1)

  int ggT(int a, int b)
  {
    int x = a;
    int y = b;
    while (y != 0)
    {
      int r = x%y;
      x = y;
      y = r;
    }
    return x; // ggT(a,b)==x;
  }

  Bruch kuerzeBruch(Bruch q)
  {
    int t = ggT(q.z, q.n);
    if (t == 0) return q;
    Bruch result = { q.z / t,q.n / t };
    return result;
  };

  inline Bruch operator+(const Bruch& p, const Bruch& q)
  {
    Bruch result = { p.z*q.n + q.z*p.n , p.n*q.n };
    return kuerzeBruch(result);
  };

  inline Bruch operator-(const Bruch& p, const Bruch& q)
  {
    Bruch result = { p.z*q.n - q.z*p.n , p.n*q.n };
    return kuerzeBruch(result);
  };

  inline Bruch operator*(const Bruch& p, const Bruch& q)
  {
    Bruch result = { p.z*q.z,p.n*q.n };
    return kuerzeBruch(result);
  };

  inline Bruch operator/(const Bruch& p, const Bruch& q)
  {
    Bruch result = { p.z*q.n,p.n*q.z };
    return kuerzeBruch(result);
  };


  // d) Zum Initialisieren:

  Bruch initBruch(int z = 1, int n = 1) // mit einer Klasse Bruch: Konstruktor
  { // InitBruch(x): x/1
    return { z,n };
  }

  std::string to_string(Bruch q)
  {
    return std::to_string(q.z) + "/" + std::to_string(q.n);
  };

    void zeigeBruchrechnung(TextBox^ tb)
  {
    Bruch b1 = initBruch(1, 2);
    Bruch b2 = initBruch(3, 4);

    std::string r_plus = to_string(b1) + "+" + to_string(b2) + "=" + to_string(b1 + b2);
    std::string r_minus = to_string(b1) + "-" + to_string(b2) + "=" + to_string(b1 - b2);
    std::string r_mult = to_string(b1) + "*" + to_string(b2) + "=" + to_string(b1*b2);
    std::string r_div = to_string(b1) + "/" + to_string(b2) + "=" + to_string(b1 / b2);

    
    tb->AppendText(rk1::to_String(r_plus + "\r\n"));
    tb->AppendText(rk1::to_String(r_minus + "\r\n"));
    tb->AppendText(rk1::to_String(r_mult + "\r\n"));
    tb->AppendText(rk1::to_String(r_div + "\r\n"));
  }

  // d)
  Bruch summierePotenzen(Bruch p, int n)
  { // summiere die Potenzen: 1 + q + q^2 ... + q^n
    Bruch s = initBruch(1), q = initBruch(1);
    for (int i = 1; i <= n; i++)
    {
      q = q*p;
      s = s + q;
    }
    return s;
  }

  Bruch Summenformel(Bruch p, int n)
  { // Summenformel: (p^(n+1) - 1)/(p-1)
    Bruch q = p, Eins = initBruch(1);
    for (int i = 2; i <= n + 1; i++) q = q*p;
    return (q - Eins) / (p - Eins);
  }

#ifdef SIMPLEUNITTESTS_H__ 
  // Dieses Makro wird durch ein #include "SimpleUnitTests.h" definiert
  bool Unittests_BruchOperatoren(int max_z, int max_n)
  {
    bool result = true;
    for (int z = -max_z; z < max_z; z++)
      for (int n = -max_n; n < max_n; n++)
        if (z != n && n != 0)
        {
          int N = abs(n);
          Bruch p = initBruch(z, n); // z/n
          Bruch s1 = summierePotenzen(p, N);
          Bruch s2 = Summenformel(p, N);
          // s3: setze z/n in die Summenformel ein
          using std::pow;
          Bruch s3 = kuerzeBruch(initBruch(pow(double(z), N + 1) - pow(double(n), N + 1), pow(double(n), N)*(z - n)));
          // Prüfe die Ergebnisse z.B. so
          rk1::Assert::AreEqual(s1 == s2, true,
            "s1=" + to_string(s1) + " s2=" + to_string(s2));
          //result=false;
          rk1::Assert::AreEqual(s1 == s3, true,
            "s1=" + to_string(s1) + " s2=" + to_string(s2));
          //result=false;
        }
    return result;
  }

  void BruchOperatoren_Aufruf()
  {
    Unittests_BruchOperatoren(7, 7); // größere Werte: Bereichsüberlauf
  }
#endif // #ifdef SIMPLEUNITTESTS_H__ 

// *********** Kap. 8 Aufgabe 8.3.2
  //----- Aufgabe 1 ------------------------------------------------

  std::ostream& operator<<(std::ostream& f, const Bruch& b)
  {
    return f << b.z << "|" << b.n;
  }

  std::istream& operator >> (std::istream& f, Bruch& b)
  {
    char Bruchstrich; // Bruchstrich einlesen und ignorieren.
    // Eventuell auch noch prüfen, ob tatsächlich ein Bruchstrich
    f >> b.z >> Bruchstrich >> b.n;
    return f;
  }

  std::string BruchToStr(Bruch b)
  {
    std::ostringstream o;
    o << b;
    return o.str();
  }

  Bruch StrToBruch(std::string s)
  {
    std::istringstream i(s);
    Bruch b;
    i >> b;
    return b;
  }


#ifdef SIMPLEUNITTESTS_H__ 
  // Dieses Makro wird durch ein #include "SimpleUnitTests.h" definiert  
  bool Unittests_BruchKonversion(int max_z, int max_n)
  {
    bool result = true;
    for (int z = -max_z; z < max_z; z++)
      for (int n = -max_z; n < max_n; n++)
        if (z != n && n != 0)
        {
          Bruch b1 = { z,n };
          std::string s = BruchToStr(b1);
          Bruch b2 = StrToBruch(s);
          rk1::Assert::AreEqual(b1 == b2, true,
            "b1=" + to_string(b1) + " b2=" + to_string(b2));
        }
    return result;
  }
#endif // #ifdef SIMPLEUNITTESTS_H__ 

// *********** Kap. 8 Aufgabe 8.4




  // --------------- Aufgabe 1 -------------------------------

  // a)
  void vertausche1(int& x, int& y)
  {              // if (&x == &y)              
                 //  x              y
                 //  x0             x0

    {
      x = x + y; //  x0+x0             x0+x0
      y = x - y; // (x0+x0)-(x0+x0)=0  (x0+x0)-(x0+x0)=0
      x = x - y; //  0-0=0
    }
  } //  x==0 && y==0

    // Unabhängig vom ursprünglichen Wert des Arguments
    // hat dieses nach dem Aufruf immer den Wert 0

  // b)
  // Falls die Argumente gleich sind, nichts machen:

  void vertausche1a(int& x, int& y)
  {        //  x              y
           //  x0             y0
    if (&x != &y)
    {
      x = x + y; //  x0+y0             x0+y0
      y = x - y; // (x0+y0)-(x0+y0)=0  (x0+y0)-(x0+y0)=0
      x = x - y; //  0-0=0
    }
  } //  x==y0 && y==x0

    // c)
  void vertausche(int& a, int& b)
  {  // hier werden die Argumente nur kopiert. Deswegen 
     // funktioniert sie auch mit zwei gleichen Argumenten.
    int h = a;
    a = b;
    b = h;
  }

  void Button_3_18_3Click(TextBox^ tb) // ???
  {
    int a = 17, b = 18;
    tb->AppendText("vor vertausche1(a,b): a=" + a.ToString() + " b=" +
      b.ToString() + "\r\n");
    vertausche1(a, b);
    tb->AppendText("nach vertausche1(a,b): a=" + a.ToString() + " b=" +
      b.ToString() + "\r\n");

    tb->AppendText("vor vertausche1(a,a): a=" + a.ToString() + "\r\n");
    vertausche1(a, a);
    tb->AppendText("nach vertausche1(a,a): a=" + a.ToString() + "\r\n");

    a = 17;
    b = 18;
    tb->AppendText("vor vertausche(a,b): a=" + a.ToString() + " b=" +
      b.ToString() + "\r\n");
    vertausche(a, b);
    tb->AppendText("nach vertausche(a,b): a=" + a.ToString() + " b=" +
      b.ToString() + "\r\n");
    tb->AppendText("vor vertausche(a,a): a=" + a.ToString() + "\r\n");
    vertausche(a, a);
    tb->AppendText("nach vertausche(a,a): a=" + a.ToString() + "\r\n");
  }



#ifdef SIMPLEUNITTESTS_H__ 
  // Dieses Makro wird durch ein #include "SimpleUnitTests.h" definiert
  void Alle_Unittests()
  {
    rk1::Assert::Init("Unittests_GlobOperatorfunktionen");
    Unittests_BruchOperatoren(7, 7); // größere Werte: Bereichsüberlauf
    Unittests_BruchKonversion(10, 10);
    rk1::Assert::Summary();
  }
#endif // #ifdef SIMPLEUNITTESTS_H__ 

} // end of namespace N_Ueberladene_Funktionen_und_Operatoren 


