﻿// *********** Kap. 15 Lösungen Kapitel 15. Templates

#include <string>
#include <sstream>   
#include <fstream>
#include <stdexcept> // für range_error
#include <vector>
#include <algorithm>
#include <set>
#include <map>
#include <iterator>
#include <utility>
#include <tuple>
#include <functional>


// *********** Kap. 15 Aufgabe 15.1 Funktions-Templates

namespace N_Loesungen_Funktions_Templates {
  using namespace System::Windows::Forms;

  // --------------- Aufgabe 1 -------------------------------

  template <class T>void vertausche(T& a, T& b)
  {
    T h = a;
    a = b;
    b = h;
  }

  template <typename T> bool kleiner(T x, T y)
  {
    return x < y;
  }

  template <typename T> bool kleiner(T* x, T* y)
  {
    return *x < *y;
  }

  bool kleiner(const char* x, const char* y)
  {
    return std::strcmp(x, y) < 0;
  }

  template <typename T> void AuswahlSort(T A[], int n)
  {
    for (int i = 0; i < n - 1; i++)
    {
      int iMin = i;
      for (int j = iMin + 1; j < n; j++)
        if (kleiner(A[j], A[iMin])) iMin = j;
      vertausche(A[iMin], A[i]);
    }
  }

  void test_Aufgabe_1(TextBox^ tb)
  {
    tb->AppendText("Aufgabe 1 \r\n");

    int a[5] = { 5,4,6,2,1 };
    AuswahlSort(a, 5);
    for (int i = 0; i < 5; i++)
      tb->AppendText(String::Format("int a[i]={0} \r\n", a[i]));

    int* p[5] = { new int(5),new int(4),new int(6),new int(2),new int(1) };
    AuswahlSort<int*>(p, 5);
    for (int i = 0; i < 5; i++)
      tb->AppendText(String::Format("int* a[i]={0} \r\n", *p[i]));

    const char* n[5] = { "15","14","16","02","01" };
    AuswahlSort<const char*>(n, 5);
    for (int i = 0; i < 5; i++)
      tb->AppendText(String::Format("char* a[i]={0} \r\n", rk1::to_String(n[i])));
  }


  // --------------- Aufgabe 2 --------------------------------

  template <class T>
  T max1(const T& a, const T& b)
  {
    return  a < b ? b : a;
  }

  template <class InputIterator, class Function>
  Function for_each(InputIterator first,
    InputIterator last, Function f)
  {
    while (first != last) f(*first++);
    return f;
  }

  void print(int x)
  {
    // ...
  }

  struct Print {
    int operator()(int x) {};
    // ...
  };

  // a) 
  // Der Compiler bestimmt zuerst den Datentyp der Template-Argumente 
  // und verwendet diesen dann, um das Template zu erzeugen.
  // Um Namenskonflikte zu vermeiden, werden die von max erzeugten 
  // Funktionen max2, max3 usw. genannt.
  int max2(const int& a, const int& b)
  {
    return  a < b ? b : a;
  }

  // b)
  double max3(const double& a, const double& b)
  {
    return  a < b ? b : a;
  }

  // c)
  // Da die Datentypen der Argumente für a und b verschieden sind, aber
  // die Typ-Parameter denselben Datentyp haben, kann der Compiler nicht
  // entscheiden, welchen er nehmen soll.
  // Zur Lösung könnte man entweder das Template mit zwei Template-Parametern
  // oder beim Aufruf eine explizite Spezialisierung angeben.

  // d)
  typedef std::function<void(int)> F;
  // oder
  typedef void(*F1) (int);
  // oder
  using F2 = std::function<void(int)>;
  // oder
  using F3 = void(*) (int);

  F for_each(int* first, int* last, F f)
  {
    while (first != last) f(*first++);
    return f;
  }

  // e)
  F for_each2(std::vector<int>::iterator first, std::vector<int>::iterator last, F f)
  {
    while (first != last) f(*first++);
    return f;
  }

  // f)
  Print for_each3(std::vector<int>::iterator first, std::vector<int>::iterator last, Print f)
  {
    ////   while (first != last) f(*first++);
    // Da das Argument vom Typ Print nicht aufgerufen werden kann,
    // kann der Aufruf f(*first++) nicht übersetzt werden.
    return f;
  }




  // --------------- Aufgabe 3 --------------------------------
  struct Bruch {
    int z, n; // z: Zähler, n: Nenner
  };

  inline bool operator==(const Bruch& q1, const Bruch& q2)
  {
    return q1.z * q2.n == q2.z * q1.n;
  }

  inline bool operator<(const Bruch& q1, const Bruch& q2)
  {
    return q1.z * q2.n < q2.z* q1.n;
  }

  namespace N_HP_STL {
    // Diese Funktionen sind in HP-STL Version in function.h so definiert:

    template <class T>
    inline bool operator!=(const T& x, const T& y)
    {
      return !(x == y);
    }

    template <class T>
    inline bool operator>(const T& x, const T& y)
    {
      return y < x;
    }

    template <class T>
    inline bool operator<=(const T& x, const T& y)
    {
      return !(y < x);
    }

    template <class T>
    inline bool operator>=(const T& x, const T& y)
    {
      return !(x < y);
    }

    // und in Visual Studio so:

    namespace rel_ops
    {	// nested namespace to hide relational operators from std
      template<class _Ty> inline
        bool operator!=(const _Ty& _Left, const _Ty& _Right)
      {	// test for inequality, in terms of equality
        return (!(_Left == _Right));
      }

      template<class _Ty> inline
        bool operator>(const _Ty& _Left, const _Ty& _Right)
      {	// test if _Left > _Right, in terms of operator<
        return (_Right < _Left);
      }

      template<class _Ty> inline
        bool operator<=(const _Ty& _Left, const _Ty& _Right)
      {	// test if _Left <= _Right, in terms of operator<
        return (!(_Right < _Left));
      }

      template<class _Ty> inline
        bool operator>=(const _Ty& _Left, const _Ty& _Right)
      {	// test if _Left >= _Right, in terms of operator<
        return (!(_Left < _Right));
      }
    }

    // Diese Funktionen stehen nach

    //    #include <utility>

    // im Namensbereich rel_ops zur Verfügung:
  } // end of namespace N_HP_STL{

  void test_Aufgabe_3()
  {
    using namespace std::rel_ops;

    Bruch b1 = { 1,2 }, b2 = { 3,4 };
    bool r1 = b1 < b2;
    bool r2 = b1 <= b2;
    bool r3 = b1 > b2;
    bool r4 = b1 >= b2;
    bool r5 = b1 == b2;
    bool r6 = b1 != b2;
  }

  // --------------- Aufgabe 4 --------------------------------

  namespace N_variadic_template {
#if VS_VERSION >= 2015   // Microsoft C/C++ VS 2012

  // Das geht mit und ohne constexpr.
    template <typename T1, typename T2>
    constexpr auto Max(const T1& a, const T2& b)
      // -> typename std::common_type<const T1&, const T2&>::type
    {
      return a > b ? a : b;
    }

    template <typename T1, typename T2, typename ... Args>
    constexpr auto Max(const T1& a, const T2& b, const Args& ... args)
      // -> typename std::common_type<const T1&, const T2&, const Args& ...>::type
    {
      return Max(Max(a, b), args...);
    }

    void test()
    {
      constexpr auto m1 = Max(1, 2);
      constexpr auto m2 = Max(1, 2, 3.0);
      constexpr auto m3 = Max(1, 2.0, 3, 'x', 5.0f);

      auto m4 = Max(std::string("1"), std::string("2"));
      auto m5 = Max("1", "2"); // m5 = 0x011a5a38 "1"
    }
#endif
  } // end of namespace N_variadic_template


} // namespace N_Loesungen_FuncTempl 

// *********** Kap. 15 Aufgabe 15.2 Klassen-Templates
namespace N_Klassen_Templates {

  /*
  Boost Software License - Version 1.0 - August 17th, 2003

  Permission is hereby granted, free of charge, to any person or organization
  obtaining a copy of the software and accompanying documentation covered by
  this license (the "Software") to use, reproduce, display, distribute,
  execute, and transmit the Software, and to prepare derivative works of the
  Software, and to permit third-parties to whom the Software is furnished to
  do so, all subject to the following:

  The copyright notices in the Software and this entire statement, including
  the above license grant, this restriction and the following disclaimer,
  must be included in all copies of the Software, in whole or in part, and
  all derivative works of the Software, unless such copies or derivative
  works are solely in the form of machine-executable object code generated by
  a source language processor.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
  SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
  FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
  ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
  DEALINGS IN THE SOFTWARE.

  // Copyright notice for the array example:

  / * The following code declares class array,
  * an STL container (as wrapper) for arrays of constant size.
  *
  * See
  *      http://www.boost.org/libs/array/
  * for documentation.
  *
  * The original author site is at: http://www.josuttis.com/
  *
  * (C) Copyright Nicolai M. Josuttis 2001.
  *
  * Distributed under the Boost Software License, Version 1.0. (See
  * accompanying file LICENSE_1_0.txt or copy at
  * http://www.boost.org/LICENSE_1_0.txt)
  *
  * /

  // Copyright notice for the scoped_ptr example:

  //  (C) Copyright Greg Colvin and Beman Dawes 1998, 1999.
  //  Copyright (c) 2001, 2002 Peter Dimov
  //
  //  Distributed under the Boost Software License, Version 1.0. (See
  //  accompanying file LICENSE_1_0.txt or copy at
  //  http://www.boost.org/LICENSE_1_0.txt)
  //
  //  http://www.boost.org/libs/smart_ptr/scoped_ptr.htm
  //
  */

  using namespace System::Windows::Forms;

  namespace fn1 { // Die im Folgenden verwendeten Dateinamen
    const std::string TestDir = "c:\\Test\\";
  }

  // --------------- Aufgabe 1 -------------------------------

  // In der Boost-Bibliothek (http://boost.org, boost\array.hpp)
  // ist das fixed size array "Array" im Wesentlichen
  // folgendermaßen definiert (stark vereinfacht):

  template<class T, unsigned int N>
  class Array {
  public:
    T elems[N];    // fixed-size array of elements of type T

                   // kein Konstruktor, damit ein Array wie ein C-Array
                   // initialisiert werden kann
  public:
    // type definitions
    typedef T              value_type;
    typedef T& reference;
    typedef const T& const_reference;
    typedef std::size_t    size_type;

    // operator[]
    reference operator[](size_type i)
    {
      if (i >= N) // bzw. (i<0 || i>=N) bei int i;
      {
        std::string msg = "Array index out of range: " +
          std::to_string(i) + " > " + std::to_string(N - 1);
        throw std::out_of_range(msg);
      }
      return elems[i];
    }

    // ohne diese Variante ist der Indexoperator für konstante
    // Arrays wie
    //    const Array<int, n> a={1,2,3};
    // nicht definiert:
    const_reference operator[](unsigned int i) const
    {
      if (i >= N) // bzw. (i<0 || i>=N) bei int i;
      {
        std::string  msg = "Array index out of range: " +
          std::to_string(i) + " > " + std::to_string(N - 1);
        throw std::out_of_range(msg);
      }
      return elems[i];
    }

    // size is constant
    static size_type size() { return N; }

  };


  void test_Aufgabe_1(TextBox^ tb)
  {
    const int n = 100;
    Array<int, n> a;
    for (int i = 0; i < n; i++) a[i] = i;
    int s = 0;
    for (int i = 0; i < n; i++) s = s + a[i];
    tb->AppendText(String::Format("s={0} \r\n", s));
    int x = a.size();
    // Da Array keine Konstruktoren hat, kann es wie ein C-Array
    // initialisiert werden:
    Array<int, n> ai = { 1,2,3 };

    const Array<int, n> c = { 1,2,3 };
    for (int i = 0; i < n; i++) s = s + c[i];
    tb->AppendText(String::Format("s={0} \r\n", s));
    try {
      a[n] = 0; // löst eine Exception aus
    }
    catch (std::exception& e)
    {
      tb->AppendText(String::Format("a[i]= " + rk1::to_String(e.what()) + "\r\n"));
    }
    try {
      a[-1] = 0; // löst eine Exception aus
    }
    catch (std::exception& e)
    {
      tb->AppendText(String::Format("a[i]= " + rk1::to_String(e.what()) + "\r\n"));
    }
  }



  // --------------- Aufgabe 2 -------------------------------

  template <typename T>
  struct Punkt1 {
    T x, y;
    Punkt1(const T& a, const T& b) :x(a), y(b) {}
    Punkt1(const Punkt1& p) :x(p.x), y(p.y) { };
  };

  template <typename T>
  struct Punkt2 {
    T x, y;
    Punkt2(const T& a, const T& b) :x(a), y(b) {}
    template <typename U>
    Punkt2(const Punkt2<U>& p) : x(p.x), y(p.y) { };
  };

  void test_Aufgabe_3()
  {
    Punkt1<int> p1a(1, 2);        Punkt2<int> p2a(1, 2);
    Punkt1<int> p1b = p1a;        Punkt2<int> p2b = p2a;
    // Punkt1<double> p1c=p1a; // ohne den zweiten Konstruktor nicht möglich
    Punkt2<double> p3b = p2a;
  }


  // --------------- Aufgabe 3 -------------------------------

  // In der Boost-Bibliothek (http://boost.org, boost\scoped_ptr.hpp)
  // ist das Klassen-Template scoped_ptr im Wesentlichen folgendermaßen 
  // definiert (stark vereinfacht).
  // Dieses Template entspricht der Klasse MyVerySimpleSmartPointer.

  template<class T> class scoped_ptr
  {
  private:
    T* ptr;

    scoped_ptr(scoped_ptr const&);
    scoped_ptr& operator=(scoped_ptr const&);
    // Der private Zuweisungsoperator verhindert Zuweisungen
    // In C++11 würde man das mit "=delete" machen.
  public:
    explicit scoped_ptr(T* p = 0) : ptr(p) {} // never throws

    ~scoped_ptr() // never throws
    {
      delete ptr;
    }

    T& operator*() const // never throws
    {
      return *ptr;
    }

    T* operator->() const // never throws
    {
      return ptr;
    }
  };

  void test_Aufgabe_scoped_ptr()
  {
    scoped_ptr<int> sp1(new int);
    *sp1 = 17;
    int* p2 = new int(18);
    scoped_ptr<int> sp2(p2);
    // sp1=sp2; // nicht möglich
  }

  // --------------- Aufgabe 4 -------------------------------

  template<class T>
  T Average(const T* data, int numElements)
  {
    T sum = 0;
    for (int i = 0; i < numElements; ++i)
      sum += data[i];
    return sum / numElements;
  };

  double Average(const int* data, int numElements)
  {
    double sum = 0;
    for (int i = 0; i < numElements; ++i)
      sum += data[i];
    return sum / numElements;
  };


  template<class T> struct float_trait {
    using T_float = T; // wie typedef T T_float;
  };

  template<> struct float_trait<int> {
    using T_float = double; // wie typedef double T_float;
  };

  template<> struct float_trait<char> {
    using T_float = double; // wie typedef double T_float;
  };

  template<class T>
  typename float_trait<T>::T_float average(const T* data,
    int numElements)
  {
    typename float_trait<T>::T_float sum = 0;
    for (int i = 0; i < numElements; ++i)
      sum += data[i];
    return sum / numElements;
  }

  /*
  Die float_trait-Variante ist nicht so leicht verständlich,
  erspart aber doppelten Code (insbesondere, falls noch weitere
  Varianten notwendig sind).
  */

  void test_Aufgabe_Average(TextBox^ tb)
  {
    const int Max = 2;
    int d1[Max] = { 1,2 };
    double a1 = Average(d1, Max);
    double d2[Max] = { 1,2 };
    double a2 = Average(d2, Max);

    double a5 = average(d1, Max);
    double a6 = average(d2, Max);

    tb->AppendText(String::Format("int d1[Max]={{1,2}}: Average={0}\r\n", a1));
    tb->AppendText(String::Format("double d2[Max]={{1,2}}: Average={0}\r\n", a2));
    tb->AppendText(String::Format("int d1[Max]={{1,2}}: average={0}\r\n", a5));
    tb->AppendText(String::Format("double d2[Max]={{1,2}}: average={0}\r\n", a6));
  }



// *********** Kap. 15 Aufgabe 15.3.2 type traits und static#assert

  // Indem man in die Datei eine Compilezeit-Konstante aufnimmt

  class MyLib {
  public: const static int version = 1;
  };

  // und diese dann in der Datei prüft, die diese verwendet:

  void test_static_assert()
  {
    static_assert(MyLib::version > 0, "MyLib::version > 0 required");
  }

} // end of namespace N_ClassTempl {


// *********** Kap. 15 Aufgabe 15.4.1 Typ-Inferenz ??? C++11/14 raus, das ist schon in core features
namespace N_Typ_Inferenz {
  using namespace System::Windows::Forms;
  // --------------- Aufgabe 1 -------------------------------

  int fii(int x) { return 0; }

  void Typ_Inferenz_Aufgaben()
  {
    // 1.

    int i = 17;
    int* pi = &i;
    const int ci = 19;
    const int* cpi = &i;

    // auto    a;      // Fehler: Intialisierer fehlt
    auto    b = &i; // Datentyp von b: int*
                    // b = 19;         // Fehler: Zuweisung von int an int*
    auto    c(i);   // Datentyp von c: int
    c = 20;         // das geht
    auto    d{ i };// Datentyp von d: int
#if VS_VERSION >= 2015   // Microsoft C/C++ VS 2015
    d = 21;         // das geht
#endif
    auto e = fii(0);// Datentyp von e: int
    auto f = fii;   // Datentyp von f: int(*)(int)
                    // f = 23;         // Fehler
    f(24);          // das geht
    auto g = new int(25); // Datentyp von g: int*
                          // g = 26;
    auto h = ci;    // Datentyp von g: int
    h = 27;         // das geht, h ist nicht const
    const auto  j = i;    // Datentyp von j: const int
                          // constexpr auto k = j; // Fehler: j ist keine Compilezeit-Konstante

// --------------- Aufgabe 2 -------------------------------
    {
      typedef char MeinTyp; // char soll eventuell zu wchar_t geändert werden
      MeinTyp v1 = 'x';
      MeinTyp v2;
    }
    { // Mit auto erspart man das typedef
      char v1 = 'x';
      auto v2 = v1;
    }
  }
} // end of namespace N_Typ_Inferenz

// *********** Kap. 15 Aufgabe 15.5 Variadische Templates
namespace N_variadische_Templates {
  using namespace System::Windows::Forms;

  template<typename T>
  auto Summe1(T v)
  {
    return v;
  }

  template<typename T, typename... Args>
  auto Summe1(T first, Args... args)
  {
    return first + Summe1(args...);
  }


  auto Summe2()
  {
    return 0;
  }

  /* oder
     template<typename T> // damit geht dann auch s2b
     auto Summe2()
     {
       return T{};
     }
  */
  template<typename T, typename... Args>
  auto Summe2(T first, Args... args)
  {
    return first + Summe2(args...);
  }

  void test_Aufgaben()
  {
    auto s1a = Summe1(1, 2, 3.0); // Lösung: s1a=6.0
    // auto s1b = Summe1("1", "2", "3"); // Fehler: + nicht für char* definiert

    std::string s1 = "1", s2 = "2", s3 = "3";
    auto s1c = Summe1(s1, s2, s3); // s1c="abc"



  // auto s1d = Summe1(); // Fehler: kann nicht ohne Argument aufgerufen werden


    auto s2a = Summe2(1, 2, 3.0); // Lösung: s2a=6.0

    // auto s2b = Summe2(s1, s2, s3); // Fehler: Keine Abbruchbedingung für string
    auto s2c = Summe2(); // int s2c=0; 
  }

} // end of namespace N_Typ_Inferenz


